@@ -31,22 +31,8 @@
 #define PY_SSIZE_T_CLEAN
 #include <Python.h>
 
-#include "config.h"
 #include "private.h"
 
-
-/* PycairoContext_FromContext
- * Create a new PycairoContext from a cairo_t
- * ctx  - a cairo_t to 'wrap' into a Python object.
- *        It is unreferenced if the PycairoContext creation fails, or if
- *        the cairo_t has an error status.
- * type - a pointer to the type to instantiate.
- *        It can be &PycairoContext_Type, or a PycairoContext_Type subtype.
- *        (cairo.Context or a cairo.Context subclass)
- * base - the base object used to create the context, or NULL.
- *        it is referenced to keep it alive while the cairo_t is being used
- * Return value: New reference or NULL on failure
- */
 PyObject *
 PycairoContext_FromContext(cairo_t *ctx, PyTypeObject *type, PyObject *base) {
   PyObject *o;
@@ -88,6 +74,48 @@ pycairo_new (PyTypeObject *type, PyObject *args, PyObject *kwds) {
   return PycairoContext_FromContext (cairo_create (s->surface), type, NULL);
 }
 
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 15, 10)
+static PyObject *
+pycairo_tag_begin (PycairoContext *o, PyObject *args) {
+    const char *tag_name;
+    const char *attributes;
+
+    if (!PyArg_ParseTuple (args,
+            "eses:Context.tag_begin", "utf-8",
+            &tag_name, "utf-8", &attributes))
+        return NULL;
+
+    Py_BEGIN_ALLOW_THREADS;
+    cairo_tag_begin (o->ctx, tag_name, attributes);
+    Py_END_ALLOW_THREADS;
+
+    PyMem_Free((void *)tag_name);
+    PyMem_Free((void *)attributes);
+
+    RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
+    Py_RETURN_NONE;
+}
+
+static PyObject *
+pycairo_tag_end (PycairoContext *o, PyObject *args) {
+    const char *tag_name;
+
+    if (!PyArg_ParseTuple (args,
+            "es:Context.tag_end", "utf-8",
+            &tag_name))
+        return NULL;
+
+      Py_BEGIN_ALLOW_THREADS;
+      cairo_tag_end (o->ctx, tag_name);
+      Py_END_ALLOW_THREADS;
+
+      PyMem_Free((void *)tag_name);
+
+      RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
+      Py_RETURN_NONE;
+}
+#endif
+
 static PyObject *
 pycairo_append_path (PycairoContext *o, PyObject *args) {
   PycairoPath *p;
@@ -130,7 +158,7 @@ pycairo_arc_negative (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_clip (PycairoContext *o) {
+pycairo_clip (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_clip (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -139,7 +167,7 @@ pycairo_clip (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_clip_extents (PycairoContext *o) {
+pycairo_clip_extents (PycairoContext *o, PyObject *ignored) {
   double x1, y1, x2, y2;
   cairo_clip_extents (o->ctx, &x1, &y1, &x2, &y2);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -147,7 +175,7 @@ pycairo_clip_extents (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_clip_preserve (PycairoContext *o) {
+pycairo_clip_preserve (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_clip_preserve (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -171,7 +199,7 @@ pycairo_in_clip (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_close_path (PycairoContext *o) {
+pycairo_close_path (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_close_path (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -180,28 +208,36 @@ pycairo_close_path (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_copy_clip_rectangle_list (PycairoContext *o) {
+pycairo_copy_clip_rectangle_list (PycairoContext *o, PyObject *ignored) {
   int i;
   PyObject *rv = NULL;
+  PyObject *rect = NULL;
   cairo_rectangle_t *r;
+
   cairo_rectangle_list_t *rlist = cairo_copy_clip_rectangle_list (o->ctx);
   if (rlist->status != CAIRO_STATUS_SUCCESS) {
     Pycairo_Check_Status (rlist->status);
     goto exit;
   }
 
-  rv = PyTuple_New(rlist->num_rectangles);
+  rv = PyList_New(rlist->num_rectangles);
   if (rv == NULL)
     goto exit;
 
   for (i = 0, r = rlist->rectangles; i < rlist->num_rectangles; i++, r++) {
     PyObject *py_rect = Py_BuildValue("(dddd)", r->x, r->y,
-				      r->width, r->height);
+                                      r->width, r->height);
     if (py_rect == NULL) {
       Py_CLEAR(rv);
       goto exit;
     }
-    PyTuple_SET_ITEM (rv, i, py_rect);
+    rect = PyObject_Call((PyObject *)&PycairoRectangle_Type, py_rect, NULL);
+    Py_DECREF (py_rect);
+    if (rect == NULL) {
+        Py_CLEAR(rv);
+        goto exit;
+    }
+    PyList_SET_ITEM (rv, i, rect);
   }
  exit:
   cairo_rectangle_list_destroy(rlist);
@@ -209,7 +245,7 @@ pycairo_copy_clip_rectangle_list (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_copy_page (PycairoContext *o) {
+pycairo_copy_page (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_copy_page (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -218,7 +254,7 @@ pycairo_copy_page (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_copy_path (PycairoContext *o) {
+pycairo_copy_path (PycairoContext *o, PyObject *ignored) {
   cairo_path_t *cp;
   Py_BEGIN_ALLOW_THREADS;
   cp = cairo_copy_path (o->ctx);
@@ -227,7 +263,7 @@ pycairo_copy_path (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_copy_path_flat (PycairoContext *o) {
+pycairo_copy_path_flat (PycairoContext *o, PyObject *ignored) {
   cairo_path_t *cp;
   Py_BEGIN_ALLOW_THREADS;
   cp = cairo_copy_path_flat (o->ctx);
@@ -274,7 +310,7 @@ pycairo_device_to_user_distance (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_fill (PycairoContext *o) {
+pycairo_fill (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_fill (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -283,7 +319,7 @@ pycairo_fill (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_fill_extents (PycairoContext *o) {
+pycairo_fill_extents (PycairoContext *o, PyObject *ignored) {
   double x1, y1, x2, y2;
   cairo_fill_extents (o->ctx, &x1, &y1, &x2, &y2);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -291,7 +327,7 @@ pycairo_fill_extents (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_fill_preserve (PycairoContext *o) {
+pycairo_fill_preserve (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_fill_preserve (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -300,7 +336,7 @@ pycairo_fill_preserve (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_font_extents (PycairoContext *o) {
+pycairo_font_extents (PycairoContext *o, PyObject *ignored) {
   cairo_font_extents_t e;
 
   cairo_font_extents (o->ctx, &e);
@@ -310,25 +346,29 @@ pycairo_font_extents (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_get_antialias (PycairoContext *o) {
+pycairo_get_antialias (PycairoContext *o, PyObject *ignored) {
   RETURN_INT_ENUM (Antialias, cairo_get_antialias (o->ctx));
 }
 
 static PyObject *
-pycairo_get_current_point (PycairoContext *o) {
+pycairo_get_current_point (PycairoContext *o, PyObject *ignored) {
   double x, y;
   cairo_get_current_point (o->ctx, &x, &y);
   return Py_BuildValue("(dd)", x, y);
 }
 
 static PyObject *
-pycairo_get_dash (PycairoContext *o) {
+pycairo_get_dash (PycairoContext *o, PyObject *ignored) {
   double *dashes = NULL, offset;
   int count, i;
   PyObject *py_dashes = NULL, *rv = NULL;
 
   count = cairo_get_dash_count (o->ctx);
-  dashes = PyMem_Malloc (count * sizeof(double));
+  if (count < 0) {
+    PyErr_SetString (PyExc_RuntimeError, "invalid dash return");
+    return NULL;
+  }
+  dashes = PyMem_Malloc ((unsigned int)count * sizeof(double));
   if (dashes == NULL)
     return PyErr_NoMemory();
 
@@ -352,30 +392,30 @@ pycairo_get_dash (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_get_dash_count (PycairoContext *o) {
-  return PYCAIRO_PyLong_FromLong (cairo_get_dash_count (o->ctx));
+pycairo_get_dash_count (PycairoContext *o, PyObject *ignored) {
+  return PyLong_FromLong (cairo_get_dash_count (o->ctx));
 }
 
 static PyObject *
-pycairo_get_fill_rule (PycairoContext *o) {
+pycairo_get_fill_rule (PycairoContext *o, PyObject *ignored) {
   RETURN_INT_ENUM(FillRule, cairo_get_fill_rule (o->ctx));
 }
 
 static PyObject *
-pycairo_get_font_face (PycairoContext *o) {
+pycairo_get_font_face (PycairoContext *o, PyObject *ignored) {
   return PycairoFontFace_FromFontFace (
        cairo_font_face_reference (cairo_get_font_face (o->ctx)));
 }
 
 static PyObject *
-pycairo_get_font_matrix (PycairoContext *o) {
+pycairo_get_font_matrix (PycairoContext *o, PyObject *ignored) {
   cairo_matrix_t matrix;
   cairo_get_font_matrix (o->ctx, &matrix);
   return PycairoMatrix_FromMatrix (&matrix);
 }
 
 static PyObject *
-pycairo_get_font_options (PycairoContext *o) {
+pycairo_get_font_options (PycairoContext *o, PyObject *ignored) {
   cairo_font_options_t *options = cairo_font_options_create();
   cairo_get_font_options (o->ctx, options);
   /* there is no reference fn */
@@ -383,141 +423,90 @@ pycairo_get_font_options (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_get_group_target (PycairoContext *o) {
+pycairo_get_group_target (PycairoContext *o, PyObject *ignored) {
   cairo_surface_t *surface = cairo_get_group_target (o->ctx);
-  if (surface != NULL)
-    return PycairoSurface_FromSurface (cairo_surface_reference (surface),
-				       NULL);
-  Py_RETURN_NONE;
+  return PycairoSurface_FromSurface (cairo_surface_reference (surface), NULL);
 }
 
 static PyObject *
-pycairo_get_line_cap (PycairoContext *o) {
+pycairo_get_line_cap (PycairoContext *o, PyObject *ignored) {
   RETURN_INT_ENUM(LineCap, cairo_get_line_cap (o->ctx));
 }
 
 static PyObject *
-pycairo_get_line_join (PycairoContext *o) {
+pycairo_get_line_join (PycairoContext *o, PyObject *ignored) {
   RETURN_INT_ENUM(LineJoin, cairo_get_line_join (o->ctx));
 }
 
 static PyObject *
-pycairo_get_line_width (PycairoContext *o) {
+pycairo_get_line_width (PycairoContext *o, PyObject *ignored) {
   return PyFloat_FromDouble(cairo_get_line_width (o->ctx));
 }
 
 static PyObject *
-pycairo_get_matrix (PycairoContext *o) {
+pycairo_get_matrix (PycairoContext *o, PyObject *ignored) {
   cairo_matrix_t matrix;
   cairo_get_matrix (o->ctx, &matrix);
   return PycairoMatrix_FromMatrix (&matrix);
 }
 
 static PyObject *
-pycairo_get_miter_limit (PycairoContext *o) {
+pycairo_get_miter_limit (PycairoContext *o, PyObject *ignored) {
   return PyFloat_FromDouble (cairo_get_miter_limit (o->ctx));
 }
 
 static PyObject *
-pycairo_get_operator (PycairoContext *o) {
+pycairo_get_operator (PycairoContext *o, PyObject *ignored) {
   RETURN_INT_ENUM(Operator, cairo_get_operator (o->ctx));
 }
 
 static PyObject *
-pycairo_get_scaled_font (PycairoContext *o) {
+pycairo_get_scaled_font (PycairoContext *o, PyObject *ignored) {
   return PycairoScaledFont_FromScaledFont (
 	   cairo_scaled_font_reference (cairo_get_scaled_font (o->ctx)));
 }
 
 static PyObject *
-pycairo_get_source (PycairoContext *o) {
+pycairo_get_source (PycairoContext *o, PyObject *ignored) {
   return PycairoPattern_FromPattern (
 	       cairo_pattern_reference (cairo_get_source (o->ctx)), NULL);
 }
 
 static PyObject *
-pycairo_get_target (PycairoContext *o) {
+pycairo_get_target (PycairoContext *o, PyObject *ignored) {
   return PycairoSurface_FromSurface (
 	       cairo_surface_reference (cairo_get_target (o->ctx)),
 	       NULL);
 }
 
 static PyObject *
-pycairo_get_tolerance (PycairoContext *o) {
+pycairo_get_tolerance (PycairoContext *o, PyObject *ignored) {
   return PyFloat_FromDouble (cairo_get_tolerance (o->ctx));
 }
 
-/* read a Python sequence of (i,x,y) sequences
- * return cairo_glyph_t *
- *        num_glyphs
- *        must call PyMem_Free(glyphs) when finished using the glyphs
- */
-static cairo_glyph_t *
-_PyGlyphs_AsGlyphs (PyObject *py_object, int *num_glyphs)
-{
-  int length, i;
-  cairo_glyph_t *glyphs = NULL, *glyph;
-  PyObject *py_glyphs, *py_seq = NULL;
-
-  py_glyphs = PySequence_Fast (py_object, "glyphs must be a sequence");
-  if (py_glyphs == NULL)
-    return NULL;
-
-  length = PySequence_Fast_GET_SIZE(py_glyphs);
-  if (*num_glyphs < 0 || *num_glyphs > length)
-    *num_glyphs = length;
-
-  glyphs = PyMem_Malloc (*num_glyphs * sizeof(cairo_glyph_t));
-  if (glyphs == NULL) {
-    PyErr_NoMemory();
-    goto error;
-  }
-  for (i = 0, glyph = glyphs; i < *num_glyphs; i++, glyph++) {
-    PyObject *py_item = PySequence_Fast_GET_ITEM(py_glyphs, i);
-    py_seq = PySequence_Fast (py_item, "glyph items must be a sequence");
-    if (py_seq == NULL)
-      goto error;
-    if (PySequence_Fast_GET_SIZE(py_seq) != 3) {
-      PyErr_SetString(PyExc_ValueError,
-		      "each glyph item must be an (i,x,y) sequence");
-      goto error;
-    }
-    glyph->index = PYCAIRO_PyLong_AsLong(PySequence_Fast_GET_ITEM(py_seq, 0));
-    glyph->x = PyFloat_AsDouble(PySequence_Fast_GET_ITEM(py_seq, 1));
-    glyph->y = PyFloat_AsDouble(PySequence_Fast_GET_ITEM(py_seq, 2));
-    if (PyErr_Occurred())
-      goto error;
-    Py_DECREF(py_seq);
-  }
-  Py_DECREF(py_glyphs);
-  return glyphs;
- error:
-  Py_DECREF(py_glyphs);
-  Py_XDECREF(py_seq);
-  PyMem_Free(glyphs);
-  return NULL;
-}
-
 static PyObject *
 pycairo_glyph_extents (PycairoContext *o, PyObject *args) {
   int num_glyphs = -1;
   cairo_glyph_t *glyphs;
   cairo_text_extents_t extents;
-  PyObject *py_object;
+  PyObject *py_object, *ext_args, *res;
 
   if (!PyArg_ParseTuple (args, "O|i:Context.glyph_extents",
 			 &py_object, &num_glyphs))
     return NULL;
 
-  glyphs = _PyGlyphs_AsGlyphs (py_object, &num_glyphs);
+  glyphs = _PycairoGlyphs_AsGlyphs (py_object, &num_glyphs);
   if (glyphs == NULL)
     return NULL;
   cairo_glyph_extents (o->ctx, glyphs, num_glyphs, &extents);
   PyMem_Free (glyphs);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
-  return Py_BuildValue("(dddddd)", extents.x_bearing, extents.y_bearing,
-		       extents.width, extents.height, extents.x_advance,
-		       extents.y_advance);
+  ext_args = Py_BuildValue ("(dddddd)", extents.x_bearing, extents.y_bearing,
+                            extents.width, extents.height, extents.x_advance,
+                            extents.y_advance);
+  res = PyObject_Call ((PyObject *)&PycairoTextExtents_Type, ext_args, NULL);
+  Py_DECREF (ext_args);
+  return res;
 }
 
 static PyObject *
@@ -530,7 +519,7 @@ pycairo_glyph_path (PycairoContext *o, PyObject *args) {
 			 &py_object, &num_glyphs))
     return NULL;
 
-  glyphs = _PyGlyphs_AsGlyphs (py_object, &num_glyphs);
+  glyphs = _PycairoGlyphs_AsGlyphs (py_object, &num_glyphs);
   if (glyphs == NULL)
     return NULL;
   cairo_glyph_path (o->ctx, glyphs, num_glyphs);
@@ -540,7 +529,7 @@ pycairo_glyph_path (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_has_current_point (PycairoContext *o) {
+pycairo_has_current_point (PycairoContext *o, PyObject *ignored) {
   PyObject *b = cairo_has_current_point (o->ctx) ? Py_True : Py_False;
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_INCREF(b);
@@ -548,7 +537,7 @@ pycairo_has_current_point (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_identity_matrix (PycairoContext *o) {
+pycairo_identity_matrix (PycairoContext *o, PyObject *ignored) {
   cairo_identity_matrix (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -637,21 +626,21 @@ pycairo_move_to (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_new_path (PycairoContext *o) {
+pycairo_new_path (PycairoContext *o, PyObject *ignored) {
   cairo_new_path (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
 }
 
 static PyObject *
-pycairo_new_sub_path (PycairoContext *o) {
+pycairo_new_sub_path (PycairoContext *o, PyObject *ignored) {
   cairo_new_sub_path (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
 }
 
 static PyObject *
-pycairo_paint (PycairoContext *o) {
+pycairo_paint (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_paint (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -674,7 +663,7 @@ pycairo_paint_with_alpha (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_path_extents (PycairoContext *o) {
+pycairo_path_extents (PycairoContext *o, PyObject *ignored) {
   double x1, y1, x2, y2;
   cairo_path_extents (o->ctx, &x1, &y1, &x2, &y2);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -682,18 +671,18 @@ pycairo_path_extents (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_pop_group (PycairoContext *o) {
+pycairo_pop_group (PycairoContext *o, PyObject *ignored) {
   return PycairoPattern_FromPattern (cairo_pop_group (o->ctx), NULL);
 }
 
 static PyObject *
-pycairo_pop_group_to_source (PycairoContext *o) {
+pycairo_pop_group_to_source (PycairoContext *o, PyObject *ignored) {
   cairo_pop_group_to_source (o->ctx);
   Py_RETURN_NONE;
 }
 
 static PyObject *
-pycairo_push_group (PycairoContext *o) {
+pycairo_push_group (PycairoContext *o, PyObject *ignored) {
   cairo_push_group (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -702,10 +691,14 @@ pycairo_push_group (PycairoContext *o) {
 static PyObject *
 pycairo_push_group_with_content (PycairoContext *o, PyObject *args) {
   cairo_content_t content;
+  int content_arg;
 
-  if (!PyArg_ParseTuple(args, "i:Context.push_group_with_content",
-			&content))
+  if (!PyArg_ParseTuple (args, "i:Context.push_group_with_content",
+                         &content_arg))
     return NULL;
+
+  content = (cairo_content_t)content_arg;
+
   cairo_push_group_with_content (o->ctx, content);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -762,14 +755,14 @@ pycairo_rel_move_to (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_reset_clip (PycairoContext *o) {
+pycairo_reset_clip (PycairoContext *o, PyObject *ignored) {
   cairo_reset_clip (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
 }
 
 static PyObject *
-pycairo_restore (PycairoContext *o) {
+pycairo_restore (PycairoContext *o, PyObject *ignored) {
   cairo_restore (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -788,7 +781,7 @@ pycairo_rotate (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_save (PycairoContext *o) {
+pycairo_save (PycairoContext *o, PyObject *ignored) {
   cairo_save (o->ctx);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -809,13 +802,18 @@ pycairo_scale (PycairoContext *o, PyObject *args) {
 static PyObject *
 pycairo_select_font_face (PycairoContext *o, PyObject *args) {
   const char *utf8;
-  cairo_font_slant_t slant   = CAIRO_FONT_SLANT_NORMAL;
-  cairo_font_weight_t weight = CAIRO_FONT_WEIGHT_NORMAL;
+  cairo_font_slant_t slant;
+  cairo_font_weight_t weight;
+  int slant_arg = CAIRO_FONT_SLANT_NORMAL;
+  int weight_arg = CAIRO_FONT_WEIGHT_NORMAL;
 
-  if (!PyArg_ParseTuple (args, PYCAIRO_ENC_TEXT_FORMAT "|ii:Context.select_font_face",
-			 "utf-8", &utf8, &slant, &weight))
+  if (!PyArg_ParseTuple (args, "es|ii:Context.select_font_face",
+                         "utf-8", &utf8, &slant_arg, &weight_arg))
     return NULL;
 
+  slant = (cairo_font_slant_t)slant_arg;
+  weight = (cairo_font_weight_t)weight_arg;
+
   cairo_select_font_face (o->ctx, utf8, slant, weight);
   PyMem_Free((void *)utf8);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -824,11 +822,14 @@ pycairo_select_font_face (PycairoContext *o, PyObject *args) {
 
 static PyObject *
 pycairo_set_antialias (PycairoContext *o, PyObject *args) {
-  cairo_antialias_t antialias = CAIRO_ANTIALIAS_DEFAULT;
+  cairo_antialias_t antialias;
+  int antialias_arg = CAIRO_ANTIALIAS_DEFAULT;
 
-  if (!PyArg_ParseTuple(args, "|i:Context.set_antialias", &antialias))
+  if (!PyArg_ParseTuple (args, "|i:Context.set_antialias", &antialias_arg))
     return NULL;
 
+  antialias = (cairo_antialias_t)antialias_arg;
+
   cairo_set_antialias (o->ctx, antialias);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -837,7 +838,7 @@ pycairo_set_antialias (PycairoContext *o, PyObject *args) {
 static PyObject *
 pycairo_set_dash (PycairoContext *o, PyObject *args) {
   double *dashes, offset = 0;
-  int num_dashes, i;
+  Py_ssize_t num_dashes, i;
   PyObject *py_dashes;
 
   if (!PyArg_ParseTuple (args, "O|d:Context.set_dash", &py_dashes, &offset))
@@ -849,7 +850,12 @@ pycairo_set_dash (PycairoContext *o, PyObject *args) {
     return NULL;
 
   num_dashes = PySequence_Fast_GET_SIZE(py_dashes);
-  dashes = PyMem_Malloc (num_dashes * sizeof(double));
+  if (num_dashes > INT_MAX) {
+      Py_DECREF (py_dashes);
+      PyErr_SetString (PyExc_ValueError, "dash sequence too large");
+      return NULL;
+  }
+  dashes = PyMem_Malloc ((unsigned int)num_dashes * sizeof(double));
   if (dashes == NULL) {
     Py_DECREF(py_dashes);
     return PyErr_NoMemory();
@@ -863,7 +869,7 @@ pycairo_set_dash (PycairoContext *o, PyObject *args) {
       return NULL;
     }
   }
-  cairo_set_dash (o->ctx, dashes, num_dashes, offset);
+  cairo_set_dash (o->ctx, dashes, (int)num_dashes, offset);
   PyMem_Free (dashes);
   Py_DECREF(py_dashes);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -873,10 +879,13 @@ pycairo_set_dash (PycairoContext *o, PyObject *args) {
 static PyObject *
 pycairo_set_fill_rule (PycairoContext *o, PyObject *args) {
   cairo_fill_rule_t fill_rule;
+  int fill_rule_arg;
 
-  if (!PyArg_ParseTuple (args, "i:Context.set_fill_rule", &fill_rule))
+  if (!PyArg_ParseTuple (args, "i:Context.set_fill_rule", &fill_rule_arg))
     return NULL;
 
+  fill_rule = (cairo_fill_rule_t)fill_rule_arg;
+
   cairo_set_fill_rule (o->ctx, fill_rule);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -939,10 +948,13 @@ pycairo_set_font_size (PycairoContext *o, PyObject *args) {
 static PyObject *
 pycairo_set_line_cap (PycairoContext *o, PyObject *args) {
   cairo_line_cap_t line_cap;
+  int line_cap_arg;
 
-  if (!PyArg_ParseTuple (args, "i:Context.set_line_cap", &line_cap))
+  if (!PyArg_ParseTuple (args, "i:Context.set_line_cap", &line_cap_arg))
     return NULL;
 
+  line_cap = (cairo_line_cap_t)line_cap_arg;
+
   cairo_set_line_cap (o->ctx, line_cap);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -951,10 +963,13 @@ pycairo_set_line_cap (PycairoContext *o, PyObject *args) {
 static PyObject *
 pycairo_set_line_join (PycairoContext *o, PyObject *args) {
   cairo_line_join_t line_join;
+  int line_join_arg;
 
-  if (!PyArg_ParseTuple (args, "i:Context.set_line_join", &line_join))
+  if (!PyArg_ParseTuple (args, "i:Context.set_line_join", &line_join_arg))
     return NULL;
 
+  line_join = (cairo_line_join_t)line_join_arg;
+
   cairo_set_line_join (o->ctx, line_join);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
@@ -999,12 +1014,15 @@ pycairo_set_miter_limit (PycairoContext *o, PyObject *args) {
 
 static PyObject *
 pycairo_set_operator(PycairoContext *o, PyObject *args) {
-  cairo_operator_t op;
+  cairo_operator_t operator;
+  int operator_arg;
 
-  if (!PyArg_ParseTuple(args, "i:Context.set_operator", &op))
+  if (!PyArg_ParseTuple (args, "i:Context.set_operator", &operator_arg))
     return NULL;
 
-  cairo_set_operator(o->ctx, op);
+  operator = (cairo_operator_t)operator_arg;
+
+  cairo_set_operator(o->ctx, operator);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
   Py_RETURN_NONE;
 }
@@ -1095,7 +1113,7 @@ pycairo_show_glyphs (PycairoContext *o, PyObject *args) {
 			 &py_object, &num_glyphs))
     return NULL;
 
-  glyphs = _PyGlyphs_AsGlyphs (py_object, &num_glyphs);
+  glyphs = _PycairoGlyphs_AsGlyphs (py_object, &num_glyphs);
   if (glyphs == NULL)
     return NULL;
   Py_BEGIN_ALLOW_THREADS;
@@ -1107,7 +1125,7 @@ pycairo_show_glyphs (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_show_page (PycairoContext *o) {
+pycairo_show_page (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_show_page (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -1119,7 +1137,7 @@ static PyObject *
 pycairo_show_text (PycairoContext *o, PyObject *args) {
   const char *utf8;
 
-  if (!PyArg_ParseTuple (args, PYCAIRO_ENC_TEXT_FORMAT ":Context.show_text", "utf-8", &utf8))
+  if (!PyArg_ParseTuple (args, "es:Context.show_text", "utf-8", &utf8))
     return NULL;
 
   Py_BEGIN_ALLOW_THREADS;
@@ -1132,7 +1150,7 @@ pycairo_show_text (PycairoContext *o, PyObject *args) {
 }
 
 static PyObject *
-pycairo_stroke (PycairoContext *o) {
+pycairo_stroke (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_stroke (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -1141,7 +1159,7 @@ pycairo_stroke (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_stroke_extents (PycairoContext *o) {
+pycairo_stroke_extents (PycairoContext *o, PyObject *ignored) {
   double x1, y1, x2, y2;
   cairo_stroke_extents (o->ctx, &x1, &y1, &x2, &y2);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
@@ -1149,7 +1167,7 @@ pycairo_stroke_extents (PycairoContext *o) {
 }
 
 static PyObject *
-pycairo_stroke_preserve (PycairoContext *o) {
+pycairo_stroke_preserve (PycairoContext *o, PyObject *ignored) {
   Py_BEGIN_ALLOW_THREADS;
   cairo_stroke_preserve (o->ctx);
   Py_END_ALLOW_THREADS;
@@ -1161,23 +1179,27 @@ static PyObject *
 pycairo_text_extents (PycairoContext *o, PyObject *args) {
   cairo_text_extents_t extents;
   const char *utf8;
+  PyObject *ext_args, *res;
 
-  if (!PyArg_ParseTuple (args, PYCAIRO_ENC_TEXT_FORMAT ":Context.text_extents", "utf-8", &utf8))
+  if (!PyArg_ParseTuple (args, "es:Context.text_extents", "utf-8", &utf8))
     return NULL;
 
   cairo_text_extents (o->ctx, utf8, &extents);
   PyMem_Free((void *)utf8);
   RETURN_NULL_IF_CAIRO_CONTEXT_ERROR(o->ctx);
-  return Py_BuildValue("(dddddd)", extents.x_bearing, extents.y_bearing,
-		       extents.width, extents.height, extents.x_advance,
-		       extents.y_advance);
+  ext_args =  Py_BuildValue ("(dddddd)", extents.x_bearing, extents.y_bearing,
+                             extents.width, extents.height, extents.x_advance,
+                             extents.y_advance);
+  res = PyObject_Call ((PyObject *)&PycairoTextExtents_Type, ext_args, NULL);
+  Py_DECREF (ext_args);
+  return res;
 }
 
 static PyObject *
 pycairo_text_path (PycairoContext *o, PyObject *args) {
   const char *utf8;
 
-  if (!PyArg_ParseTuple (args, PYCAIRO_ENC_TEXT_FORMAT ":Context.text_path", "utf-8", &utf8))
+  if (!PyArg_ParseTuple (args, "es:Context.text_path", "utf-8", &utf8))
     return NULL;
 
   cairo_text_path (o->ctx, utf8);
@@ -1236,6 +1258,87 @@ pycairo_user_to_device_distance (PycairoContext *o, PyObject *args) {
   return Py_BuildValue("(dd)", dx, dy);
 }
 
+static PyObject *
+pycairo_show_text_glyphs (PycairoContext *o, PyObject *args) {
+  const char *utf8 = NULL;
+  PyObject *glyphs_arg, *glyphs_seq = NULL;
+  PyObject *clusters_arg, *clusters_seq = NULL;
+  cairo_text_cluster_flags_t cluster_flags;
+  int cluster_flags_arg;
+  PyObject *py_item;
+  cairo_glyph_t *glyphs = NULL;
+  cairo_text_cluster_t *clusters = NULL;
+  Py_ssize_t i, clusters_size, glyphs_size;
+
+  if (!PyArg_ParseTuple (args,
+      "esOOi:Context.show_text_glyphs",
+      "utf-8", &utf8, &glyphs_arg, &clusters_arg, &cluster_flags_arg))
+    return NULL;
+
+  cluster_flags = (cairo_text_cluster_flags_t)cluster_flags_arg;
+
+  glyphs_seq = PySequence_Fast (glyphs_arg, "glyphs must be a sequence");
+  if (glyphs_seq == NULL)
+    goto error;
+  glyphs_size = PySequence_Fast_GET_SIZE (glyphs_seq);
+  if (glyphs_size > INT_MAX) {
+    PyErr_SetString (PyExc_ValueError, "glyph sequence too large");
+    goto error;
+  }
+  glyphs = cairo_glyph_allocate ((int)glyphs_size);
+  if (glyphs_size && glyphs == NULL) {
+    PyErr_NoMemory();
+    goto error;
+  }
+  for (i=0; i < glyphs_size; i++) {
+    py_item = PySequence_Fast_GET_ITEM (glyphs_seq, i);
+    if (py_item == 0 || _PyGlyph_AsGlyph (py_item, &(glyphs[i])) != 0)
+      goto error;
+  }
+  Py_CLEAR (glyphs_seq);
+
+  clusters_seq = PySequence_Fast (clusters_arg, "clusters must be a sequence");
+  if (clusters_seq == NULL)
+    goto error;
+  clusters_size = PySequence_Fast_GET_SIZE (clusters_seq);
+  if (clusters_size > INT_MAX) {
+    PyErr_SetString (PyExc_ValueError, "clusters sequence too large");
+    goto error;
+  }
+  clusters = cairo_text_cluster_allocate ((int)clusters_size);
+  if (clusters_size && clusters == NULL) {
+    PyErr_NoMemory();
+    goto error;
+  }
+  for (i=0; i < clusters_size; i++) {
+    py_item = PySequence_Fast_GET_ITEM (clusters_seq, i);
+    if (py_item == NULL ||
+        _PyTextCluster_AsTextCluster (py_item, &(clusters[i])) != 0)
+      goto error;
+  }
+  Py_CLEAR (clusters_seq);
+
+  Py_BEGIN_ALLOW_THREADS;
+  cairo_show_text_glyphs (
+    o->ctx, utf8, -1, glyphs, (int)glyphs_size, clusters,
+    (int)clusters_size, cluster_flags);
+  Py_END_ALLOW_THREADS;
+
+  PyMem_Free ((void *)utf8);
+  utf8 = NULL;
+  cairo_glyph_free (glyphs);
+  cairo_text_cluster_free (clusters);
+
+  RETURN_NULL_IF_CAIRO_CONTEXT_ERROR (o->ctx);
+  Py_RETURN_NONE;
+error:
+  PyMem_Free ((void *)utf8);
+  cairo_glyph_free (glyphs);
+  cairo_text_cluster_free (clusters);
+  Py_XDECREF (glyphs_seq);
+  Py_XDECREF (clusters_seq);
+  return NULL;
+}
 
 static PyMethodDef pycairo_methods[] = {
   /* methods never exposed in a language binding:
@@ -1248,6 +1351,10 @@ static PyMethodDef pycairo_methods[] = {
    * - not needed since Pycairo calls Pycairo_Check_Status() to check
    *   for errors and raise exceptions
    */
+#if CAIRO_VERSION >= CAIRO_VERSION_ENCODE(1, 15, 10)
+  {"tag_begin",       (PyCFunction)pycairo_tag_begin,        METH_VARARGS},
+  {"tag_end",         (PyCFunction)pycairo_tag_end,          METH_VARARGS},
+#endif
   {"append_path",     (PyCFunction)pycairo_append_path,      METH_VARARGS},
   {"arc",             (PyCFunction)pycairo_arc,              METH_VARARGS},
   {"arc_negative",    (PyCFunction)pycairo_arc_negative,     METH_VARARGS},
@@ -1351,9 +1458,30 @@ static PyMethodDef pycairo_methods[] = {
   {"user_to_device",  (PyCFunction)pycairo_user_to_device,   METH_VARARGS},
   {"user_to_device_distance",(PyCFunction)pycairo_user_to_device_distance,
    METH_VARARGS},
+  {"show_text_glyphs",(PyCFunction)pycairo_show_text_glyphs, METH_VARARGS},
   {NULL, NULL, 0, NULL},
 };
 
+static PyObject*
+pycairo_richcompare (PyObject *self, PyObject *other, int op)
+{
+  if (Py_TYPE(self) == Py_TYPE(other))
+    return Pycairo_richcompare (
+      ((PycairoContext *)self)->ctx,
+      ((PycairoContext *)other)->ctx,
+      op);
+  else {
+    Py_INCREF(Py_NotImplemented);
+    return Py_NotImplemented;
+  }
+}
+
+static Py_hash_t
+pycairo_hash (PyObject *self)
+{
+  return PYCAIRO_Py_hash_t_FromVoidPtr (((PycairoContext *)self)->ctx);
+}
+
 PyTypeObject PycairoContext_Type = {
   PyVarObject_HEAD_INIT(NULL, 0)
   "cairo.Context",                    /* tp_name */
@@ -1368,7 +1496,7 @@ PyTypeObject PycairoContext_Type = {
   0,                                  /* tp_as_number */
   0,                                  /* tp_as_sequence */
   0,                                  /* tp_as_mapping */
-  0,                                  /* tp_hash */
+  pycairo_hash,                       /* tp_hash */
   0,                                  /* tp_call */
   0,                                  /* tp_str */
   0,                                  /* tp_getattro */
@@ -1378,7 +1506,7 @@ PyTypeObject PycairoContext_Type = {
   0,                                  /* tp_doc */
   0,                                  /* tp_traverse */
   0,                                  /* tp_clear */
-  0,                                  /* tp_richcompare */
+  pycairo_richcompare,                /* tp_richcompare */
   0,                                  /* tp_weaklistoffset */
   0,                                  /* tp_iter */
   0,                                  /* tp_iternext */
